Attribute VB_Name = "modMMC"
Option Explicit

Public cd As New clsCDROM                             ' CD-ROM class

Public Type t_MSFLBA
    M                           As Byte               ' minutes
    s                           As Byte               ' seconds
    F                           As Byte               ' frames
    LBA                         As Long               ' logical block address
End Type

Public Type t_CDInfo
    Capacity                    As Long               ' capacity (only CD-R[W])
    LeadIn                      As t_MSFLBA           ' Lead-In Start
    LeadOut                     As t_MSFLBA           ' Lead-Out Start
    DiscStatus                  As e_Status           ' CD Status
    LastSessionStatus           As e_Status           ' last session's status
    CDType                      As e_CDType           ' CD Type
    CDSubType                   As e_CD_SubType       ' Sub Type
    Erasable                    As Boolean            ' erasable?
    Tracks                      As Byte               ' number of tracks
    Sessions                    As Byte               ' number of sessions
    Size                        As Double               ' size of the disc
    vendor                      As String             ' CD-R(W) vendor
End Type

Public Type t_ATIP
    Length(1)                   As Byte               ' data len
    rsvd1(1)                    As Byte               ' reserved
    ITWP                        As Byte               ' ?
    Uru                         As Byte               ' ?
    DiscType                    As Byte               ' CD Type (CD-R/CD-RW)
    rsvd2                       As Byte               ' reserved
    LeadIn_Min                  As Byte               ' Lead-In Start (minutes)
    LeadIn_Sec                  As Byte               ' Lead-In Start (seconds)
    LeadIn_Frm                  As Byte               ' Lead-In Start (frames)
    rsvd3                       As Byte               ' reserved
    LeadOut_Min                 As Byte               ' Lead-Out Start (minutes)
    LeadOut_Sec                 As Byte               ' Lead-Out Start (seconds)
    LeadOut_Frm                 As Byte               ' Lead-Out Start (frames)
    rsvd4                       As Byte               ' reserved
    Rest(12)                    As Byte               ' rest
End Type

Public Type t_RDI
    PageLen(1)                  As Byte               ' Page len
    states                      As Byte               ' misc. data (erasable, ...)
    FirstTrack                  As Byte               ' first track on the disc (experience: 1...)
    NumSessionsLSB              As Byte               ' number of sessions
    FirstTrackLastSessionLSB    As Byte               ' first track in last session
    LastTrackLastSessionLSB     As Byte               ' last track in last session
    misc                        As Byte               ' misc.
    DiscType                    As Byte               ' CD sub type (CD-ROM/CD-I/XA)
    NumSessionsMSB              As Byte               ' number of sessions
    FirstTrackLastSessionMSB    As Byte               ' first track in last session
    LastTrackLastSessionMSB     As Byte               ' last track in last session
    DiscIdentification(3)       As Byte               ' CD ID
    LastSessionLeadInStart(3)   As Byte               ' Lead-In start time (h:m:s:f)
    LastPossibleLeadOutStart(3) As Byte               ' last possible Lead-Out Start (h:m:s:f)
    DBC(6)                      As Byte               ' Disc Bar Code
End Type

Public Type t_ReadCap
    Blocks(3)                   As Byte               ' written sectors
    BlockLen(3)                 As Byte               ' sectorsize
End Type

Public Type t_Track
    TrkNum                      As Byte               ' track number
    AudioTrack                  As Boolean            ' Audio Track?
    start                       As t_MSFLBA           ' Track Start
    end                         As t_MSFLBA           ' Track Ende
End Type

Public Type t_Tracks
    Tracks                      As Byte               ' number of tracks
    Track(98)                   As t_Track            ' Track Collection (max. 99)
End Type

Public Type t_TOC_TRACK
    rsvd1                       As Byte               ' reserved
    ADR                         As Byte               ' ADR
    Track                       As Byte               ' Track
    rsvd2                       As Byte               ' reserved
    addr(3)                     As Byte               ' address (h:m:s:f or LBA)
End Type

Public Type t_TOC_STRUCT
    TocLen(1)                   As Byte               ' data len
    FirstTrack                  As Byte               ' first track
    LastTrack                   As Byte               ' letzter track
    TocTrack(99)                As t_TOC_TRACK        ' tracks
End Type

Public Type t_TrackInfo
    Track                       As Byte               ' track number
    Session                     As Byte               ' session number
    DataMode                    As e_SectorModes      ' data mode
    startLBA                    As Long               ' Start LBA
    endLBA                      As Long               ' End LBA
    Length                      As Long               ' length in sectors
    LastTrackInSession          As Boolean            ' last track in session?
End Type

Private Type t_RTOC_ENTRY
    sessionNr                   As Byte               ' current session
    ADRCTL                      As Byte               ' ADR/CTRL
    TNO                         As Byte               ' track number
    point                       As Byte               ' packet type
    min                         As Byte               ' minutes
    sec                         As Byte               ' seconds
    frm                         As Byte               ' frames
    zero                        As Byte               ' 0
    pmin                        As Byte               ' Point minutes
    psec                        As Byte               ' Point seconds
    pframe                      As Byte               ' Point frames
End Type

Public Type t_RTOC_STRUCT
    dummy(3)                    As Byte               ' Header
    packet(255)                 As t_RTOC_ENTRY       ' packets
End Type

Public Type t_DVD_CPYINFO
    Length(1)                   As Byte
    rsvd(1)                     As Byte
    cpyprotectsystype           As Byte
    regioninfo                  As Byte
End Type

Public Type t_DVD_Phys
    Length(1)                   As Byte               ' data len
    rsvd(1)                     As Byte               ' reserved
    BookType                    As Byte               ' Booktype (DVD-ROM, DVD-RAM, DVD-R, DVD+R, ...)
    discsize                    As Byte               ' DVD size/Max. Rate
    LayerType                   As Byte               ' number of layers/Layer Type
    TrackDens                   As Byte               ' Track density
    zero                        As Byte               ' 0
    StartSector(2)              As Byte               ' physical start sector
    zero2                       As Byte               ' 0
    EndSector(2)                As Byte               ' physical end sector
    zero3                       As Byte               ' 0
    EndSectorLayer0(2)          As Byte               ' physical end sector in layer 0
    bca                         As Byte               ' bca
    mspec(2030)                 As Byte               ' depends on disc
End Type

Public Type t_MMC
    PageCode                    As Byte               ' Page Code
    PageLen                     As Byte               ' Page len
    rsvd2(7)                    As Byte               ' reserved
    ReadSupported               As Byte               ' readable formats
    WriteSupported              As Byte               ' writable formats
    misc(3)                     As Byte               ' misc.
    MaxReadSpeed(1)             As Byte               ' max. read speed
    NumVolLevels(1)             As Byte               ' num. volume levels
    BufferSize(1)               As Byte               ' buffer size
    CurrReadSpeed(1)            As Byte               ' curr. read speed
    rsvd                        As Byte               ' reserved
    misc2                       As Byte               ' misc.
    MaxWriteSpeed(1)            As Byte               ' max. write speed
    CurrWriteSpeed(1)           As Byte               ' curr write speed
    RotationControl             As Byte
    CurrWriteSpeedMMC3(1)       As Byte
End Type

Public Type t_Speed
    MaxRSpeed                   As Integer            ' max. read speed
    MaxWSpeed                   As Integer            ' max. write speed
    CurrRSpeed                  As Integer            ' curr. read speed
    CurrWSpeed                  As Integer            ' curr write speed
End Type

Public Type t_ReadFeatures                            ' can read:
    CDR                         As Boolean            '      CD-R
    CDRW                        As Boolean            '      CD-RW
    DVDR                        As Boolean            '      DVD-R
    DVDROM                      As Boolean            '      DVD-ROM
    DVDRAM                      As Boolean            '      DVD-RAM
    subchannels                 As Boolean            '      Sub-Channels
    SubChannelsCorrected        As Boolean            '      Sub-Channels corrected
    SubChannelsFormLeadIn       As Boolean            '      Sub-Channels from Lead-In
    C2ErrorPointers             As Boolean            '      C2 Error Pointers
    ISRC                        As Boolean            '      International Standard Recording Code
    UPC                         As Boolean            '      ?
    BC                          As Boolean            '      Bar Code
    Mode2Form1                  As Boolean            '      Mode 2 Form 1 sectors
    Mode2Form2                  As Boolean            '      Mode 2 Form 2
    Multisession                As Boolean            '      Multi-Session CDs
    CDDARawRead                 As Boolean            '      Audio sectors
End Type

Public Type t_WriteModes
    Raw96                       As Boolean            ' Raw + 96
    Raw16                       As Boolean            ' Raw + 16
    SAO                         As Boolean            ' Session At Once
    TAO                         As Boolean            ' Track At Once
    Raw96Test                   As Boolean            ' Raw + 96 + Test-Mode
    Raw16Test                   As Boolean            ' Raw + 16 + Test-Mode
    SAOTest                     As Boolean            ' Session At Once + Test-Mode
    TAOTest                     As Boolean            ' Track At Once + Test-Mode
End Type

Public Type t_WriteFeatures                           ' can write:
    CDR                         As Boolean            '      CD-R
    CDRW                        As Boolean            '      CD-RW
    DVDR                        As Boolean            '      DVD-R
    DVDRAM                      As Boolean            '      DVD-RAM
    TestMode                    As Boolean            '      Test-Mode
    BURNProof                   As Boolean            '      BURN-Proof
    WriteModes                  As t_WriteModes       ' supported write modes
End Type

Public Type t_DrvInfo
    ReadFeatures                As t_ReadFeatures     ' read features
    WriteFeatures               As t_WriteFeatures    ' write features
    speeds                      As t_Speed            ' speeds
    AnalogAudio                 As Boolean            ' analog audio playback?
    JitterCorrection            As Boolean            ' jitter effect correction?
    BufferSize                  As Long               ' buffer size
    LockMedia                   As Boolean            ' can lock media?
    LoadingMechanism            As e_LoadingMechanism ' loading mechanism
    Interface                   As e_DrvInterfaces    ' drive interface
End Type

Public Type t_RTI
    DataLen(1)                  As Byte               ' data length
    TrackNumLSB                 As Byte               ' track number
    SessionNumLSB               As Byte               ' track session
    rsvd                        As Byte               ' reserved
    TrackMode                   As Byte               ' Track Mode
    DataMode                    As Byte               ' Data Mode
    misc                        As Byte               ' misc.
    Track_Start(3)              As Byte               ' track start (h:m:s:f)
    Track_Next_Writable(3)      As Byte               ' next writable address (h:m:s:f)
    Track_Free_Blocks(3)        As Byte               ' free sectors (h:m:s:f)
    Track_Packet_Size(3)        As Byte               ' Fixed Packet size (?)
    Track_Size(3)               As Byte               ' Track size (LBA)
    Track_Last_Recorded(3)      As Byte               ' last written address
    TrackNumMSB                 As Byte               ' Track Number
    SessionNumMSB               As Byte               ' Track Session
    rsvd2(1)                    As Byte               ' reserved
End Type

Private Type t_InqDat
    PDT                         As Byte               ' drive type
    PDQ                         As Byte               ' removable drive
    VER                         As Byte               ' MMC Version (zero for ATAPI)
    RDF                         As Byte               ' interface depending field
    DLEN                        As Byte               ' additional len
    rsv1(1)                     As Byte               ' reserved
    Feat                        As Byte               ' ?
    VID(7)                      As Byte               ' vendor
    PID(15)                     As Byte               ' Product
    PVER(3)                     As Byte               ' revision (= Firmware Version)
    FWVER(20)                   As Byte               ' ?
End Type

Public Type t_BufferCapacity
    DataLen(1)                  As Byte               ' data len
    rsvd(1)                     As Byte               ' reserved
    bufferlen(3)                As Byte               ' buffer size
    BufferBlank(3)              As Byte               ' empty part of the buffer
End Type

Public Type t_WavHdr
    riff                        As String * 4         ' RIFF Header
    len                         As Long               ' file length
    WavFmt                      As String * 8         ' Wave Format
    HdrLen                      As Long               ' header length
    format                      As Integer            ' format
    NumChannels                 As Integer            ' channels
    SampleRate                  As Long               ' frequency
    BytesPerSec                 As Long               ' Bytes/second
    BlockAlign                  As Integer            ' Block Align
    BitsPerSample               As Integer            ' Bits/Sample
    Data                        As String * 4         ' Data Chunk
    DataLen                     As Long               ' datalength w/o header
End Type

'Structure of the Q Sub-Channel data header returned by ReadSubChannel
Public Type t_SubQHeader
    rsvd As Byte            ' reserved
    audio_stat As Byte      ' audio status (Play, Stop, ...)
    Data_Len(1) As Byte     ' data length
End Type

'Structure of the Q Sub-Channel information (Current Position) returned by ReadSubChannel
Public Type t_CDCurrPos
    header As t_SubQHeader  ' header (top)
    dataformat As Byte      ' data format of the requested information
    ADRCTRL As Byte         ' ADR/CTRL
    TrkNum As Byte          ' track number
    IndexNum As Byte        ' index number of the current track (0 = Pre-Gap, 1 = data)
    AbsCDAddr(3) As Byte    ' absolute address (position)
    TrkRelAddr(3) As Byte   ' track relative address (position within a track)
End Type

'Structure of the Q Sub-Channel information (ISRC) returned by ReadSubChannel
Public Type t_Q_ISRC
    header As t_SubQHeader  ' header (top)
    dataformat As Byte      ' data format of the requested information
    ADRCTRL As Byte         ' ADR/CTRL
    TrkNum As Byte          ' track number
    rsvd As Byte            ' reserved
    ISRC(15) As Byte        ' ISRC (International Standard Recording Code)
End Type

'Structure of the Q Sub-Channel information (MCN) returned by ReadSubChannel
Public Type t_Q_MCN
    header As t_SubQHeader  ' header (top)
    dataformat As Byte      ' data format of the requested information
    rsvd(2) As Byte         ' reserved
    MCN(15) As Byte         ' MCN (Media Catalog Number)
End Type

Public Type t_Feat_Hdr
    DataLen(3) As Byte      ' data length
    rsvd(1) As Byte         ' reserved
    curr_profile(1) As Byte ' current profile
End Type

Public Type t_Feat_CD_READ
    hdr As t_Feat_Hdr       ' header
    code(1) As Byte         ' Feature code
    VER As Byte             ' version (= 1)
    additional_len As Byte  ' additional length (= 4)
    CDText As Byte          ' can read CD-Text and C2?
    rsvd(2) As Byte         ' reserved
End Type

Public Type t_Feat_DVD_P_RW
    hdr As t_Feat_Hdr       ' header
    code(1) As Byte         ' Feature code
    VER As Byte             ' version (= 0)
    additional_len As Byte  ' additional length (= 4)
    write As Byte           ' can write DVD+RW
    close_only As Byte      ' background format
    rsvd(1) As Byte         ' reserved
End Type

'Feature 2Fh: DVD-R/RW
Public Type t_Feat_DVD_R_RW
    hdr As t_Feat_Hdr       ' header
    code(1) As Byte         ' Feature code
    VER As Byte             ' version (= 1)
    additional_len As Byte  ' additional length (= 4)
    writeDVDRW As Byte      ' can write DVD-RW
    rsvd(2) As Byte         ' reserved
End Type

'Feature 2Bh: DVD+R
Public Type t_Feat_DVD_P_R
    hdr As t_Feat_Hdr       ' header
    code(1) As Byte         ' Feature code
    VER As Byte             ' version
    additional_len As Byte  ' additional length
    write As Byte           ' can write DVD+R
    rsvd(2) As Byte         ' reserved
End Type

' Feature 3Bh: DVD+R Double Layer
Public Type t_Feat_DVD_P_R_DL
    hdr As t_Feat_Hdr       ' header
    code(1) As Byte         ' Feature code
    VER As Byte             ' version
    additional_len As Byte  ' additional length
    write As Byte           ' can write DVD+R DL
    rsvd(2) As Byte         ' reserved
End Type

' Feature 28h: Mount Rainer
Public Type t_Feat_MRW
    hdr As t_Feat_Hdr       ' header
    code(1) As Byte         ' Feature code
    VER As Byte             ' version
    additional_len As Byte  ' additional length
    write As Byte           ' can write MRW
    rsvd(2) As Byte         ' reserved
End Type

' Multimedia Capabilities page write speed descriptor
Public Type t_MMCP_WriteSpeed
    rsvd        As Byte     ' reserved
    rotation    As Byte     ' rotation control
    speed(1)    As Byte     ' speed in kb/s
End Type

' generic Q Sub-Channel format
Public Type t_Qch
    CTLADR      As Byte     ' control/ADR
    DATAQ(8)    As Byte     ' Q data
    CRC(1)      As Byte     ' CRC
End Type

' Mode 1 Q channel (0001b)
Public Type t_Qch_Mode1
    CTLADR      As Byte     ' control/ADR
    TNO         As Byte     ' track number
    index       As Byte     ' track index
    min         As Byte     ' rel. minutes
    sec         As Byte     ' rel. seconds
    frm         As Byte     ' rel. frames
    zero        As Byte     ' zero
    amin        As Byte     ' abs. minutes
    asec        As Byte     ' abs. seconds
    afrm        As Byte     ' abs. frames
    CRC(1)      As Byte     ' CRC
End Type

' Mode 2 Q channel (0010b)
Public Type t_Qch_Mode2
    CTLADR      As Byte     ' control/ADR
    MCN(7)      As Byte     ' media catalog number
    afrm        As Byte     ' abs. frames
    CRC(1)      As Byte     ' CRC
End Type

' Mode 3 Q channel (0011b)
Public Type t_Qch_Mode3
    CTLADR      As Byte     ' control/ADR
    ISRC(7)     As Byte     ' 6 bit cells
    afrm        As Byte     ' abs. frames
    CRC(1)      As Byte     ' CRC
End Type

Public Enum e_DrvInterfaces
    IF_SCSI                                           ' SCSI
    IF_ATAPI                                          ' ATAPI
    IF_IEEE                                           ' IEEE
    IF_USB                                            ' USB
    IF_UNKNWN                                         ' unknown
End Enum

Public Enum e_READCD_FLAGS
    RCD_SYNC = &H80                                   ' sync pattern
    RCD_HDR_4BT = &H20                                ' 4-Bytes Header
    RCD_HDR_8BT = &H40                                ' 8-Bytes Header
    RCD_USRDATA = &H10                                ' userdata
    RCD_EDC_ECC = &H8                                 ' EDC+ECC correction
    RCD_RAW = &HF8                                    ' full sector
End Enum

Public Enum e_READCD_SUBCH_FLAGS
    RCD_SUBCH_PW_RAW = &H1                            ' Raw Channels P-W
    RCD_SUBCH_Q = &H2                                 ' only Q channel
    RCD_SUBCH_PW_CORRECTED = &H4                      ' Channels P-W corrected
End Enum

Public Enum e_BLANKTYPE
    BLANK_FULL                                        ' full blank
    BLANK_QUICK                                       ' quick blank
End Enum

Public Enum e_SectorModes
    MODE_AUDIO                                        ' Audio (or Mode-0)
    MODE_MODE1                                        ' Mode-1
    MODE_MODE2                                        ' Plain Mode-2
    MODE_MODE2_FORM1                                  ' Mode-2 Form-1
    MODE_MODE2_FORM2                                  ' Mode-2 Form-2
End Enum

Public Enum e_Genres
    GENRE_ADULT_CONTEMP = 4                           ' adult contemporary
    GENRE_ALT_ROCK = 5                                ' alternative rock
    GENRE_CHILDREN = 6                                ' children music
    GENRE_CLASSIC = 7                                 ' classic
    GENRE_CHRIST_CONTEMP = 6                          ' christian contemporary
    GENRE_COUNTRY = 7                                 ' country
    GENRE_DANCE = 8                                   ' dance
    GENRE_EASY_LISTENING = 9                          ' easy listening
    GENRE_EROTIC = 10                                 ' erotic
    GENRE_FOLK = 11                                   ' folk
    GENRE_GOSPEL = 12                                 ' gospel
    GENRE_HIPHOP = 13                                 ' hip hop
    GENRE_JAZZ = 14                                   ' jazz
    GENRE_LATIN = 15                                  ' latin
    GENRE_MUSICAL = 16                                ' musical
    GENRE_NEWAGE = 17                                 ' new age
    GENRE_OPERA = 18                                  ' opera
    GENRE_OPERETTA = 19                               ' operetta
    GENRE_POP = 20                                    ' pop
    GENRE_RAP = 21                                    ' rap
    GENRE_REGGAE = 22                                 ' reggae
    GENRE_ROCK = 23                                   ' rock
    GENRE_RYTHMANDBLUES = 24                          ' R'n'B
    GENRE_SOUNDEFFECTS = 25                           ' sound effects
    GENRE_SPOKEN_WORD = 26                            ' spoken words
    GENRE_WORLD_MUSIC = 27                            ' world music
End Enum

Public Enum e_CDType
    ROMTYPE_CDROM                                     ' CD-ROM
    ROMTYPE_CDR                                       ' CD-R
    ROMTYPE_CDRW                                      ' CD-RW
    ROMTYPE_CDROM_R_RW                                ' CD-ROM, CD-R oder CD-RW
    ROMTYPE_DVD_ROM                                   ' DVD-ROM
    ROMTYPE_DVD_R                                     ' DVD-R
    ROMTYPE_DVD_RW                                    ' DVD-RW
    ROMTYPE_DVD_RAM                                   ' DVD-RAM
    ROMTYPE_DVD_P_R                                   ' DVD+R
    ROMTYPE_DVD_P_RW                                  ' DVD+RW
End Enum

Public Enum e_LoadingMechanism
    LOAD_CADDY                                        ' Caddy
    LOAD_TRAY                                         ' Tray
    LOAD_POPUP                                        ' Popup
    LOAD_CHANGER                                      ' Changer
    LOAD_UNKNWN                                       ' Unknown
End Enum

Public Enum e_Status
    STAT_EMPTY                                        ' empty
    STAT_INCOMPLETE                                   ' uncomplete
    STAT_COMPLETE                                     ' complete
    STAT_UNKNWN                                       ' unknown
End Enum

Public Enum e_CD_SubType
    STYPE_CDROMDA                                     ' CD-ROM or CDDA
    STYPE_CDI                                         ' CD-I
    STYPE_XA                                          ' CD-XA
    STYPE_UNKNWN                                      ' unknown
End Enum

'(from CDROM-TOOL [GPL])
Public Enum e_SpinDown
    SD_VS                                             ' vendor specific
    SD_125MS                                          ' 125 ms
    SD_250MS                                          ' 250     "
    SD_500MS                                          ' 500     "
    SD_1SEC                                           '   1 sec
    SD_2SEC                                           '   2     "
    SD_4SEC                                           '   4     "
    SD_8SEC                                           '   8     "
    SD_16SEC                                          '  16     "
    SD_32SEC                                          '  32     "
    SD_1MIN                                           '   1 min
    SD_2MIN                                           '   2     "
    SD_4MIN                                           '   4     "
    SD_8MIN                                           '   8     "
    SD_16MIN                                          '  16     "
    SD_32MIN                                          '  32     "
End Enum

Public Enum e_CDTextPacketTypes
    CD_TEXT_PACK_ALBUM_NAME = &H80
    CD_TEXT_PACK_PERFORMER = &H81
    CD_TEXT_PACK_SONGWRITER = &H82
    CD_TEXT_PACK_COMPOSER = &H83
    CD_TEXT_PACK_ARRANGER = &H84
    CD_TEXT_PACK_MESSAGES = &H85
    CD_TEXT_PACK_DISC_ID = &H86
    CD_TEXT_PACK_GENRE = &H87
    CD_TEXT_PACK_TOC_INFO = &H88
    CD_TEXT_PACK_TOC_INFO2 = &H89
    CD_TEXT_PACK_UPC_EAN = &H8E
    CD_TEXT_PACK_SIZE_INFO = &H8F
End Enum

'datasectorsynchronisationpattern :)
Public Const SYNCPATTERN As String = "00FFFFFFFFFFFFFFFFFFFF00"

Public Function CDRomTestUnitReady(ByVal DrvID As String) As Boolean

    Dim cmd(5) As Byte
    Dim i      As Integer

    If cd.ExecCMD(DrvID, cmd, 6, False, SRB_DIR_IN, 0, 0) Then
        CDRomTestUnitReady = True
        Exit Function
    End If

    ' no disc present
    If cd.LastASC = &H3A Then
        '
        Exit Function

    ' unit is becoming ready,
    ' or not ready to ready change
    ' because medium may have changed,
    ' wait for it
    ElseIf (cd.LastASC = &H4 And cd.LastASCQ = &H1) _
    Or (cd.LastSK = 6 And cd.LastASC = 40) Then

        ' try 5 times (~5 seconds)
        For i = 1 To 5
            If cd.ExecCMD(DrvID, cmd, 6, False, SRB_DIR_IN, 0, 0) Then Exit For
            Sleep 1000
        Next i

        CDRomTestUnitReady = cd.ExecCMD(DrvID, cmd, 6, False, SRB_DIR_IN, 0, 0)

    End If

End Function

'disc present?
Public Function CDRomIsDiscPresent(ByVal DrvID As String) As Boolean

    Dim media_event_req(8) As Byte

    'get Tray Status
    If Not CDRomGetEventStatusNotification(DrvID, &H10, _
                                           VarPtr(media_event_req(0)), _
                                           UBound(media_event_req)) Then

        CDRomIsDiscPresent = CDRomTestUnitReady(DrvID)
        Exit Function

    End If

    'valid data?
    If media_event_req(0) = 0 And media_event_req(1) = 0 Then
        CDRomIsDiscPresent = CDRomTestUnitReady(DrvID)
        Exit Function
    Else
        'disc present?
        CDRomIsDiscPresent = IsBitSet(media_event_req(5), 1)
    End If

End Function

'Media locked?
Public Function CDRomIsTrayLocked(ByVal DrvID As String) As Boolean
    Dim mmc As t_MMC

    'read MM Capabilities Page
    If Not CDRomModeSense10(DrvID, &H2A, VarPtr(mmc), Len(mmc) - 1, True, True) Then _
        CDRomIsTrayLocked = -1: Exit Function

    'check "Lock State" Bit
    CDRomIsTrayLocked = Abs(IsBitSet(mmc.misc(2), 1))
End Function

'Tray open or closed?
Public Function CDRomIsTrayOpen(ByVal DrvID As String) As Long

    Dim media_event_req(8) As Byte

    If Not CDRomTestUnitReady(DrvID) Then

        If cd.LastASC = &H3A And cd.LastASCQ = &H1 Then
            CDRomIsTrayOpen = False
        ElseIf cd.LastASC = &H3A And cd.LastASCQ = &H2 Then
            CDRomIsTrayOpen = True

        ' drive doesn't report door status with TUR
        ElseIf cd.LastASC = &H3A And cd.LastASCQ = 0 Then

            'get Tray-Status
            If Not CDRomGetEventStatusNotification(DrvID, &H10, _
                                                   VarPtr(media_event_req(0)), _
                                                   UBound(media_event_req)) Then

                CDRomIsTrayOpen = -1: Exit Function

            End If

            'valid data?
            If media_event_req(0) = 0 And media_event_req(1) = 0 Then
                CDRomIsTrayOpen = -1: Exit Function
            Else
                'Tray open or closed?
                CDRomIsTrayOpen = Abs(IsBitSet(media_event_req(5), 0))
            End If

        End If

    Else

        CDRomIsTrayOpen = False

    End If

End Function

'read a Mode Page
Public Function CDRomModeSense10(ByVal DrvID As String, ByVal MP As Byte, _
                                 ByVal PtrBuffer As Long, _
                                 ByVal bufferlen As Long, _
                                 Optional ByVal DBD As Boolean, _
                                 Optional ByVal CV As Boolean) As Boolean

    Dim cmd(9) As Byte

    cmd(0) = &H5A                       ' MODE SENSE 10 OpCode
    cmd(1) = Abs(DBD) * &H8             ' Disable Block Descriptors
    cmd(2) = MP Or (Abs(Not CV) * &H80) ' Mode Page (default values)
    cmd(7) = cd.RShift(bufferlen, 8)    ' allocation length
    cmd(8) = bufferlen And &HFF         ' allocation length

    CDRomModeSense10 = cd.ExecCMD(DrvID, cmd, 10, False, _
                                  SRB_DIR_IN, PtrBuffer, bufferlen + 1)

End Function

'get Event/Status
Public Function CDRomGetEventStatusNotification(ByVal DrvID As String, _
                                                ByVal Request As Byte, _
                                                ByVal PtrBuffer As Long, _
                                                ByVal bufferlen As Long) As Boolean

    Dim cmd(9) As Byte

    cmd(0) = &H4A                   ' GET EVENT/STATUS NOTIFICATION OpCode
    cmd(1) = 1                      '
    cmd(4) = Request                ' Request Type
    cmd(8) = bufferlen              ' allocation length

    CDRomGetEventStatusNotification = cd.ExecCMD( _
                                       DrvID, cmd, 10, False, SRB_DIR_IN, _
                                        PtrBuffer, bufferlen _
                                      )

End Function

'get track mode by its sectorcontents
'from CDR-DAO (GPL)
Public Function CDRomGetSectorMode(ByVal DrvID As String, _
                                   ByVal SectorLBA As Long) As e_SectorModes

    Dim buffer(2351) As Byte
    Dim sync As String, cnt As Integer

    'read sector raw
    If Not CDRomReadCD(DrvID, SectorLBA, 1, VarPtr(buffer(0)), _
                       UBound(buffer), &HF8) Then

        CDRomGetSectorMode = -1
        Exit Function

    End If

    'get the first 12 bytes and convert them to hex
    For cnt = 0 To 11
        sync = sync & format(Hex$(buffer(cnt)), "00")
    Next

    'is a sync pattern?
    If sync = SYNCPATTERN Then

        'has to be a data track

        'Byte 16 is the mode
        Select Case buffer(15)

            Case 1
                'Mode-1 Track (2048 Bytes data)
                CDRomGetSectorMode = MODE_MODE1

            Case 2
                'compare pattern
                If (buffer(16) = buffer(20)) And (buffer(17) = buffer(21)) _
                    And (buffer(18) = buffer(22)) And (buffer(19) = buffer(23)) Then

                    If Not buffer(18) And &H20 = 0 Then
                        'Mode 2 Form 2 Track
                        CDRomGetSectorMode = MODE_MODE2_FORM2
                    Else
                        'Mode 2 Form 1 Track
                        CDRomGetSectorMode = MODE_MODE2_FORM1
                    End If

                Else
                    'Mode-2 Track (2336 Bytes data)
                    CDRomGetSectorMode = MODE_MODE2
                End If

        End Select

    Else
        'Audio or Mode-0 (2352 Bytes data/2352 Bytes completely empty)
        CDRomGetSectorMode = MODE_AUDIO
    End If
End Function

'read sectors
Public Function CDRomReadCD(ByVal DrvID As String, ByVal LBA As Long, _
                            ByVal NumSectors As Long, ByVal PtrBuf As Long, _
                            ByVal buflen As Long, ByVal ReadFlags As e_READCD_FLAGS, _
                            Optional ByVal SubchBits As e_READCD_SUBCH_FLAGS _
                           ) As Boolean

    Dim cmd(11) As Byte

    cmd(0) = &HBE                               ' READ CD OpCode
    cmd(2) = cd.RShift(LBA, 24) And &HFF        ' LBA
    cmd(3) = cd.RShift(LBA, 16) And &HFF
    cmd(4) = cd.RShift(LBA, 8) And &HFF
    cmd(5) = LBA And &HFF                       ' LBA
    cmd(6) = cd.RShift(NumSectors, 16) And &HFF ' num. sectors
    cmd(7) = cd.RShift(NumSectors, 8) And &HFF
    cmd(8) = NumSectors And &HFF                ' num. sectors
    cmd(9) = ReadFlags                          ' read flags
    cmd(10) = SubchBits                         ' Sub-Channel Flags

    CDRomReadCD = cd.ExecCMD(DrvID, cmd, 12, False, _
                             SRB_DIR_IN, PtrBuf, buflen + 1, 40)

End Function

'read track info
Public Function CDRomTrackInfo(ByVal driveid As String, ByVal Track As Integer, _
                               ByRef nfo As t_TrackInfo) As Boolean

    Dim LeadOut As Long, LastTrk As Long
    Dim toc As t_RTOC_STRUCT
    Dim i As Integer

    'read raw TOC
    If Not CDRomReadTOC(driveid, 2, True, 1, _
                        VarPtr(toc), Len(toc) - 1) Then Exit Function

    'go through all packets
    For i = 0 To ((cd.LShift(toc.dummy(0), 8) Or toc.dummy(1)) \ 11) - 1

        'determine packet type
        Select Case toc.packet(i).point

            'found the chosen track?
            Case Is = Track

                'found the track, save its number
                nfo.Track = Track
                'Sessionnumber
                nfo.Session = toc.packet(i).sessionNr

                'Startaddress
                nfo.startLBA = cd.MSF2LBA(toc.packet(i).pmin, _
                                          toc.packet(i).psec, _
                                          toc.packet(i).pframe)

                'last track in the current session?
                If toc.packet(i).point = LastTrk Then

                    'yes, last LBA is Lead-Out startaddress
                    nfo.endLBA = LeadOut - 150
                    nfo.LastTrackInSession = True

                Else

                    'else normal end address
                    nfo.endLBA = cd.MSF2LBA(toc.packet(i + 1).pmin, _
                                            toc.packet(i + 1).psec, _
                                            toc.packet(i + 1).pframe)

                End If

                'tracklength in sectors
                nfo.Length = nfo.endLBA - nfo.startLBA

                'determine data mode
                nfo.DataMode = CDRomGetSectorMode(driveid, nfo.startLBA)

                Exit For

            'found packet: "last track in current session"
            Case &HA1
                LastTrk = toc.packet(i).pmin

            'found packet: "Lead-Out of the current session"
            Case &HA2
                LeadOut = cd.MSF2LBA(toc.packet(i).pmin, _
                                     toc.packet(i).psec, _
                                     toc.packet(i).pframe)

        End Select

    Next

    CDRomTrackInfo = True
End Function

'read TOC/PMA/ATIP/CD-TEXT
Public Function CDRomReadTOC(ByVal DrvID As String, ByVal TOC_Format As Integer, _
                             ByVal MSF As Boolean, ByVal Track_Session As Integer, _
                             ByVal PtrBuffer As Long, ByVal bufferlen As Long _
                            ) As Boolean

    Dim cmd(9) As Byte

    cmd(0) = &H43                           ' READ TOC OpCode
    cmd(1) = IIf(MSF, &H2, 0)               ' MSF or LBA?
    cmd(2) = TOC_Format                     ' Format (TOC, PMA, ATIP, CD-Text)
    cmd(6) = Track_Session                  ' Track/Session
    cmd(7) = bufferlen \ &HFF               ' allocation length
    cmd(8) = bufferlen And &HFF             ' allocation length

    CDRomReadTOC = cd.ExecCMD(DrvID, cmd, 10, False, _
                              SRB_DIR_IN, PtrBuffer, bufferlen)

End Function

'send a Mode Page
Public Function CDRomModeSelect10(ByVal DrvID As String, ByVal PtrBuffer As Long, _
                                  ByVal bufferlen As Long) As Boolean

    Dim cmd(9) As Byte

    cmd(0) = &H55                   ' MODE SELECT10 OpCode
    cmd(1) = &H10                   ' PF = 1 (Page Format)
    cmd(7) = bufferlen \ &HFF       ' allocation length
    cmd(8) = bufferlen Mod &HFF     ' allocation length

    CDRomModeSelect10 = cd.ExecCMD(DrvID, cmd, 10, True, _
                                   SRB_DIR_OUT, PtrBuffer, bufferlen)

End Function

'collection drive information
Public Function CDRomGetLWInfo(ByVal strDrvID As String) As t_DrvInfo

    Dim mmc As t_MMC

    'read Multimedia Capabilities Mode Page
    CDRomModeSense10 strDrvID, &H2A, VarPtr(mmc), Len(mmc) - 1, True, True

    'read read features
    With CDRomGetLWInfo.ReadFeatures
        .CDR = IsBitSet(mmc.ReadSupported, 0)
        .CDRW = IsBitSet(mmc.ReadSupported, 1)
        .DVDROM = IsBitSet(mmc.ReadSupported, 3)
        .DVDR = IsBitSet(mmc.ReadSupported, 4)
        .DVDRAM = IsBitSet(mmc.ReadSupported, 5)
        .CDDARawRead = IsBitSet(mmc.misc(1), 0)
        .Mode2Form1 = IsBitSet(mmc.misc(0), 4)
        .Mode2Form2 = IsBitSet(mmc.misc(0), 5)
        .Multisession = IsBitSet(mmc.misc(0), 6)
        .ISRC = IsBitSet(mmc.misc(1), 5)
        .UPC = IsBitSet(mmc.misc(1), 6)
        .BC = IsBitSet(mmc.misc(1), 7)
        .subchannels = IsBitSet(mmc.misc(1), 2)
        .SubChannelsCorrected = IsBitSet(mmc.misc(1), 3)
        .SubChannelsFormLeadIn = IsBitSet(mmc.misc(3), 5)
        .C2ErrorPointers = IsBitSet(mmc.misc(1), 4)
    End With

End Function

'read readspeeds
Public Function CDRomGetSpeed(ByVal strDrv As String) As t_Speed
    Dim mmc As t_MMC

    CDRomModeSense10 strDrv, &H2A, VarPtr(mmc), Len(mmc) - 1, True, True

    With CDRomGetSpeed
        .MaxRSpeed = cd.LShift(mmc.MaxReadSpeed(0), 8) Or mmc.MaxReadSpeed(1)
        .CurrRSpeed = cd.LShift(mmc.CurrReadSpeed(0), 8) Or mmc.CurrReadSpeed(1)
    End With
    
End Function

'gets the interface of a drive
Public Function CDRomGetInterface(ByVal strDrv As String) As e_DrvInterfaces
    Dim buffer(15) As Byte, cmd(9) As Byte
    Dim inquiry As t_InqDat

    'try to read the drive's core feature
    CDRomGetConfiguration strDrv, 1, 2, VarPtr(buffer(0)), UBound(buffer)

    'determine the interface from it
    Select Case cd.LShift(buffer(12), 24) Or cd.LShift(buffer(13), 16) Or _
                cd.LShift(buffer(14), 8) Or buffer(15)

        Case 1: CDRomGetInterface = IF_SCSI
        Case 2, 7: CDRomGetInterface = IF_ATAPI
        Case 3, 4, 6: CDRomGetInterface = IF_IEEE
        Case 8: CDRomGetInterface = IF_USB
        Case Else: CDRomGetInterface = IF_UNKNWN

    End Select

    'if it didn't work, try INQUIRY
    If CDRomGetInterface = IF_UNKNWN Then

        cmd(0) = &H12               ' Inquiry OpCode
        cmd(4) = Len(inquiry) - 1   ' allocation length

        If cd.ExecCMD(strDrv, cmd, 10, False, SRB_DIR_IN, _
                      VarPtr(inquiry), Len(inquiry), 10) Then

            If inquiry.rsv1(0) = 0 Then CDRomGetInterface = IF_ATAPI

        End If

    End If
End Function

'read drive features
Public Function CDRomGetConfiguration(ByVal strDrv As String, ByVal StartFeature As Byte, _
                                      ByVal RT As Byte, ByVal PtrBuffer As Long, _
                                      ByVal buflen As Long) As Boolean

    Dim cmd(9) As Byte

    cmd(0) = &H46                   ' GET CONFIGURATION Op-Code
    cmd(1) = RT                     ' RT Byte
    cmd(2) = StartFeature \ &HFF    ' startfeature
    cmd(3) = StartFeature And &HFF  ' startfeature
    cmd(7) = buflen \ &HFF          ' allocation length
    cmd(8) = buflen And &HFF        ' allocation length

    CDRomGetConfiguration = cd.ExecCMD(strDrv, cmd, 10, False, _
                                       SRB_DIR_IN, PtrBuffer, buflen + 1, 10)
End Function

'collects some information about the inserted disc
Public Function CDRomGetCDInfo(ByVal strDrv As String) As t_CDInfo

    Dim BufAtip         As t_ATIP
    Dim BufRDI          As t_RDI


    'get some information
    CDRomReadDiscInformation strDrv, VarPtr(BufRDI), Len(BufRDI) - 1
    CDRomReadTOC strDrv, 4, True, 0, VarPtr(BufAtip), Len(BufAtip) - 1

    With CDRomGetCDInfo

        'CD Type
        .CDType = CDRomGetCDType(strDrv)

        'Sessions
        .Sessions = cd.LShift(BufRDI.NumSessionsMSB, 8) Or _
                              BufRDI.NumSessionsLSB

        'Tracks
        .Tracks = cd.LShift(BufRDI.LastTrackLastSessionMSB, 8) Or _
                            BufRDI.LastTrackLastSessionLSB

        'capacity in bytes (Mode 1)
        .Capacity = .LeadOut.LBA * 2048&
        If .Capacity < 0 Then .Capacity = 0

        'used part of the disc in bytes
        .Size = CDRomGetUsedBytes(strDrv)
    End With
    
End Function

'Warning: the written sectors will be multiplied with 2048.
'         Mode 2 or DA not supported.
Private Function CDRomGetUsedBytes(ByVal strDrv As String) As Double

    Dim cap     As t_ReadCap
    Dim cmd(9)  As Byte

    cmd(0) = &H25                     ' READ CAPACITY Op-Code

    If Not cd.ExecCMD(strDrv, cmd, 10, False, SRB_DIR_IN, _
                      VarPtr(cap), Len(cap), 10) Then

        'failed
        CDRomGetUsedBytes = -1

    Else

        'return written sectors
        CDRomGetUsedBytes = CDbl(cd.LShift(cap.Blocks(0), 24) Or _
                                 cd.LShift(cap.Blocks(1), 16) Or _
                                 cd.LShift(cap.Blocks(2), 8) Or _
                                 cap.Blocks(3)) _
                                * 2048#

    End If
End Function

Public Function CDRomGetIdleTimer(ByVal strDrvID As String) As Long

    Dim mpage(19) As Byte

    If CDRomModeSense10(strDrvID, &H1A, VarPtr(mpage(0)), UBound(mpage), True, True) Then

        CDRomGetIdleTimer = cd.LShift(mpage(12), 24) Or _
                            cd.LShift(mpage(13), 16) Or _
                            cd.LShift(mpage(14), 8) Or _
                            mpage(15)

    End If

End Function

Public Function CDRomGetStandbyTimer(ByVal strDrvID As String) As Long

    Dim mpage(19) As Byte

    If CDRomModeSense10(strDrvID, &H1A, VarPtr(mpage(0)), UBound(mpage), True, True) Then

        CDRomGetStandbyTimer = cd.LShift(mpage(16), 24) Or _
                            cd.LShift(mpage(17), 16) Or _
                            cd.LShift(mpage(18), 8) Or _
                            mpage(19)

    End If

End Function

'read disc information
Public Function CDRomReadDiscInformation(ByVal strDrv As String, _
                                         ByVal PtrBuffer As Long, _
                                         ByVal bufferlen As Long) As Boolean

    Dim cmd(9) As Byte

    cmd(0) = &H51                   ' READ DISC INFORMATION Op-Code
    cmd(7) = bufferlen \ &HFF       ' allocation length
    cmd(8) = bufferlen Mod &HFF     ' allocation length

    CDRomReadDiscInformation = cd.ExecCMD(strDrv, cmd, 10, False, _
                                          SRB_DIR_IN, PtrBuffer, bufferlen, 10)
End Function

'trys to get the type of the inserted CD/DVD
Public Function CDRomGetCDType(ByVal strDrv As String) As e_CDType

    Dim BufRDI          As t_RDI        ' buffer for ReadDiscInformation
    Dim BufAtip         As t_ATIP       ' ATIP for CD-R/CD-RW
    Dim conf_hdr(512)   As Byte         ' configuration header (active profile)

    'read disc information
    CDRomReadDiscInformation strDrv, VarPtr(BufRDI), Len(BufRDI) - 1

    'first try to read the ATIP to exclude CD-R/W
    If Not CDRomReadTOC(strDrv, 4, True, 0, VarPtr(BufAtip), Len(BufAtip) - 1) Then

        'if the Lead-Out is 255:255.255 MSF, it should be a CD-ROM
        If cd.MSF2LBA(BufRDI.LastPossibleLeadOutStart(1), _
                      BufRDI.LastPossibleLeadOutStart(2), _
                      BufRDI.LastPossibleLeadOutStart(3)) _
            = cd.MSF2LBA(255, 255, 255) Then

            'normal CD-ROM
            CDRomGetCDType = ROMTYPE_CDROM

        Else

            'could be a CD-ROM/R/RW
            CDRomGetCDType = ROMTYPE_CDROM_R_RW

        End If

    Else

        'ATIP could be read, either CD-R oder CD-RW.
        'but we could get fooled, so check the ATIP data :)

        'valid Lead-In start time?
        If BufAtip.LeadIn_Min > 0 Or _
           BufAtip.LeadIn_Sec > 0 Or _
           BufAtip.LeadIn_Frm > 0 Then

            'valide Lead-Out start time?
            If BufAtip.LeadIn_Min < 255 Or _
               BufAtip.LeadIn_Sec < 255 Or _
               BufAtip.LeadIn_Frm < 255 Then

                If IsBitSet(BufAtip.DiscType, 6) Then
                    'CD-RW
                    CDRomGetCDType = ROMTYPE_CDRW
                Else
                    'CD-R
                    CDRomGetCDType = ROMTYPE_CDR
                End If

            End If
        End If

    End If

    'is DVD in drive?
    If IsDVD(strDrv) Then

        'seems to be a DVD, determine its type
        CDRomGetCDType = GetDVDBookType(strDrv)

        'didn't work?
        If CDRomGetCDType = -1 Then

            'read the configuration header, we want the active profile
            If CDRomGetConfiguration(strDrv, 0, 2, VarPtr(conf_hdr(0)), _
                                     UBound(conf_hdr)) Then
    
                'CD type by the active drive profile
                Select Case (cd.LShift(conf_hdr(6), 8) Or conf_hdr(7))
                    Case &H8: CDRomGetCDType = ROMTYPE_CDROM       ' CD-ROM
                    Case &H9: CDRomGetCDType = ROMTYPE_CDR         ' CD-R
                    Case &HA: CDRomGetCDType = ROMTYPE_CDRW        ' CD-RW
                    Case &H10: CDRomGetCDType = ROMTYPE_DVD_ROM    ' DVD-ROM
                    Case &H11: CDRomGetCDType = ROMTYPE_DVD_R      ' DVD-R
                    Case &H12: CDRomGetCDType = ROMTYPE_DVD_RAM    ' DVD-RAM
                    Case &H13: CDRomGetCDType = ROMTYPE_DVD_RW     ' DVD-RW
                    Case &H14: CDRomGetCDType = ROMTYPE_DVD_RW     ' DVD-RW
                    Case &H1A: CDRomGetCDType = ROMTYPE_DVD_P_RW   ' DVD+RW
                    Case &H1B: CDRomGetCDType = ROMTYPE_DVD_P_R    ' DVD+R
                    Case Else: CDRomGetCDType = ROMTYPE_CDROM_R_RW ' das ging also mal garnicht...
                End Select
    
            End If

        End If

    End If
End Function

'simple DVD detection
Public Function IsDVD(ByVal strDrv As String) As Boolean
    Dim dummy(512) As Byte

    'DVD?
    If CDRomReadDVDStructure(strDrv, 0, 0, 0, VarPtr(dummy(0)), UBound(dummy)) Then
        If dummy(0) > 0 Or dummy(1) > 0 Then IsDVD = True
    End If
End Function

'read DVD structure
Public Function CDRomReadDVDStructure(ByVal strDrv As String, ByVal LBA As Long, _
                                      ByVal LayerNr As Byte, ByVal format As Byte, _
                                      ByVal PtrBuffer As Long, _
                                      ByVal bufferlen As Long) As Boolean

    Dim cmd(11) As Byte

    cmd(0) = &HAD                           ' READ DVD STRUCTURE Op-Code
    cmd(2) = cd.RShift(LBA, 24) And &HFF    ' LBA LSB
    cmd(3) = cd.RShift(LBA, 16) And &HFF
    cmd(4) = cd.RShift(LBA, 8) And &HFF
    cmd(5) = LBA And &HFF                   ' LBA MSB
    cmd(6) = LayerNr                        ' Layer Number
    cmd(7) = format                         ' Information Format
    cmd(8) = bufferlen \ &HFF               ' allocation length
    cmd(9) = bufferlen And &HFF             ' allocation length

    CDRomReadDVDStructure = cd.ExecCMD(strDrv, cmd, 12, False, _
                                       SRB_DIR_IN, PtrBuffer, bufferlen, 10)
End Function

'read DVD Book
Private Function GetDVDBookType(ByVal strDrv As String) As e_CDType
    Dim physdata As t_DVD_Phys
    Dim book As Byte

    'get DVD Book
    If CDRomReadDVDStructure(strDrv, 0, 0, 0, VarPtr(physdata), _
                          Len(physdata) - 1) Then

        '
        With physdata

            If IsBitSet(.BookType, 4) Then book = 1
            If IsBitSet(.BookType, 5) Then book = book Or 2
            If IsBitSet(.BookType, 6) Then book = book Or 4
            If IsBitSet(.BookType, 7) Then book = book Or 8

            Select Case book
                Case 0: GetDVDBookType = ROMTYPE_DVD_ROM   ' DVD-ROM
                Case 1: GetDVDBookType = ROMTYPE_DVD_RAM   ' DVD-RAM
                Case 2: GetDVDBookType = ROMTYPE_DVD_R     ' DVD-R
                Case 3: GetDVDBookType = ROMTYPE_DVD_RW    ' DVD-RW
                Case 9: GetDVDBookType = ROMTYPE_DVD_P_RW  ' DVD+RW
                Case 10: GetDVDBookType = ROMTYPE_DVD_P_R  ' DVD+R
                Case Else: GetDVDBookType = -1             ' unknown
            End Select

        End With

    End If
End Function

'bit in a byte is set?
Public Function IsBitSet(ByVal InByte As Byte, ByVal Bit As Byte) As Boolean
    IsBitSet = ((InByte And (2 ^ Bit)) > 0)
End Function
